import os
import boto3
import logging
import time
import redis
from typing import Optional
from dcicutils.env_utils import full_env_name
from dcicutils.function_cache_decorator import function_cache
from dcicutils.misc_utils import ignored, PRINT
from dcicutils.redis_tools import RedisBase, RedisSessionToken, SESSION_TOKEN_COOKIE
from dcicutils.redis_utils import create_redis_client
from ...app import app
from .cookie_utils import read_cookie, read_cookie_bool
from .envs import Envs
from .jwt_utils import JWT_AUDIENCE_PROPERTY_NAME, JWT_SUBJECT_PROPERTY_NAME, jwt_decode, jwt_encode
from .misc_utils import get_request_domain

logging.basicConfig()
logger = logging.getLogger(__name__)


# Constant for authtoken derived from JWT to give info about the user to
# the front-end
AUTH_TOKEN_COOKIE = 'authtoken'


class Auth:

    def __init__(self, auth0_client: str, auth0_secret: str, envs: Envs):
        self._auth0_client = auth0_client
        self._auth0_secret = auth0_secret
        self._envs = envs
        self._redis = None

    def get_redis_handler(self):
        """
        Returns a handler to Redis or None if not in use
        """
        if not self._redis:
            # 2023-09-21: Moved this from __init__ to here;
            # it speeds up provision/deploy from 4dn-cloud-infra.
            try:
                redis_url = os.environ['REDIS_HOST']
                if redis_url and ("redis://" in redis_url or "rediss://" in redis_url):
                    self._redis = RedisBase(create_redis_client(url=redis_url))
            except (redis.exceptions.ConnectionError, redis.exceptions.TimeoutError):
                PRINT('Cannot connect to Redis')
                PRINT('This error is expected when deploying with remote (ElastiCache) Redis')
        return self._redis

    @classmethod
    def get_redis_namespace(cls, env: str) -> str:
        ignored(env)
        # As of April 2023 simply use a static non-per-environment namespace for the Redis
        # auth token; this is so we can switch among different environments in Foursight
        # with the same login session, as it was working before the Redis auth work;
        return "foursight"

    def authorize(self, request: dict, env: Optional[str] = None) -> dict:
        """
        Verifies that the given request is authenticated AND authorized, based on the authtoken
        cookie (a JWT-signed-encoded value) in the given request. If so, returns the verified and
        decoded authtoken as a dictionary. If not, returns a dictionary indicating not authorized
        and/or not authenticated, and containing the basic info from the authtoken.
        """
        try:
            # Read the c4_st token (new Redis session token if Redis is in use)
            if self._redis:
                c4_st = read_cookie(request, SESSION_TOKEN_COOKIE)
                redis_session_token = RedisSessionToken.from_redis(
                    redis_handler=self._redis,
                    namespace=self.get_redis_namespace(env),
                    token=c4_st
                )
                # if this session token is not valid, nothing else is to be trusted, so bail here
                if (not redis_session_token or
                        not redis_session_token.validate_session_token(redis_handler=self._redis)):
                    return self._create_unauthenticated_response(request, "missing-or-invalid-session-token")

            # Read the authtoken cookie (will always be present).

            authtoken = read_cookie(request, AUTH_TOKEN_COOKIE)
            if not authtoken:
                return self._create_unauthenticated_response(request, "no-authtoken")

            # Decode the authtoken cookie.

            authtoken_decoded = self.decode_authtoken(authtoken)
            if not authtoken_decoded:
                return self._create_unauthenticated_response(request, "invalid-authtoken")

            # Sanity check the decoded authtoken.

            if authtoken_decoded["authenticated"] is not True:
                return self._create_unauthenticated_response(request, "invalid-authtoken-auth", authtoken_decoded)

            if self._auth0_client != authtoken_decoded[JWT_AUDIENCE_PROPERTY_NAME]:
                return self._create_unauthenticated_response(request, "invalid-authtoken-aud", authtoken_decoded)

            domain = get_request_domain(request)
            if domain != authtoken_decoded["domain"]:
                return self._create_unauthenticated_response(request, "invalid-authtoken-domain", authtoken_decoded)

            # Check the authtoken expiration time (its expiration time must be in the future i.e. greater than now).

            authtoken_expires_time_t = authtoken_decoded["authenticated_until"]
            current_time_t = int(time.time())
            if authtoken_expires_time_t <= current_time_t:
                return self._create_unauthenticated_response(request, "authtoken-expired", authtoken_decoded)

            # If given an env, then check that the specified environment is allowed, i.e. that
            # the request is AUTHORIZED. If not, return HTTP 403 (not authorized); we do NOT,
            # in this case, return 401 (not authenticated), as (above) if not authenticated;
            # the UI acts differently for these two cases. These HTTP statuses are actually
            # set via the @route decorator in react_route_decorator.

            if env:
                allowed_envs = authtoken_decoded["allowed_envs"]
                if not self._envs.is_allowed_env(env, allowed_envs):
                    status = "not-authorized-env" if self._envs.is_known_env(env) else "not-authorized-unknown-env"
                    return self._create_unauthorized_response(request, status, authtoken_decoded, is_authenticated=True)

            return {**authtoken_decoded, "authorized": True}

        except Exception as e:
            logger.error(f"Authorization exception: {e}")
            return self._create_unauthenticated_response(request, "exception: " + str(e))

    def create_authtoken(self, jwt: str, jwt_expires_at: int, domain: str, request: Optional[dict] = None) -> str:
        """
        Creates and returns a new signed JWT, to be used as the login authtoken (cookie), from
        the given AUTHENTICATED and signed and encoded JWT, which will contain the following:
        - Booleans indicating authenticated AND authorized.
        - The user name (from the "email" field of the given JWT).
        - The list of known environments.
        - The default environment.
        - The initial environment (i.e. the environment in which the user was first authenticated).
        - The list of allowed (authorized) environments for the user associated with the given JWT.
        - The first/last name of the user associated with the given JWT.
        - The timestamp of when the given JWT was issued (UTC time_t/epoch format).
        - The timestamp of when the given JWT expires issued (UTC time_t/epoch format).
        - The audience (aka "aud" aka Auth0 client ID).
        The allowed environments and first/last name are obtained via the users ElasticSearch store;
        the first/last names are just for informational/display purposes in the client.
        Returns the JWT-signed-encoded authtoken value as a string.
        """
        jwt_decoded = jwt_decode(jwt, self._auth0_client, self._auth0_secret)
        email = jwt_decoded.get("email")
        # Just FYI for UI, communicate how the user logged in, i.e. currently via either
        # Google or GitHub. For the former the "sub" property of the JWT looks something
        # like "google-oauth2|117300206013007398924", and for the latter "github|105234079".
        authenticator = jwt_decoded.get(JWT_SUBJECT_PROPERTY_NAME)
        if authenticator:
            authenticator = authenticator.casefold()
            if "google" in authenticator:
                authenticator = "google"
            elif "github" in authenticator:
                authenticator = "github"
        try:
            if request:
                # Note that these "test_mode_xyz" cookies are for testing only
                # and if used must be manually set, e.g. via Chrome Developer Tools.
                test_mode_access_key_simulate_error = read_cookie_bool(request, "test_mode_access_key_simulate_error")
                if test_mode_access_key_simulate_error:
                    # For testing only, we simulate a portal access key error (e.g. due to expiration),
                    # which would manifest itself, primarily and most importantly, here, on login.
                    raise Exception("test_mode_access_key_simulate_error")
            allowed_envs, first_name, last_name = self._envs.get_user_auth_info(email)
            user_exception = False
        except Exception as e:
            #
            # Here there was a problem getting the user info via Portal (e.g. due to expired or otherwise bad
            # Portal acesss key); this will NOT prevent the user from being logged in BUT it WILL prevent the
            # user from doing anything because there will be no allowed environments. We note this particular
            # error with a flag (user_exception) in the authtoken JWT cookie; we catch this case in the /header
            # endpoint to send back information about the portal access key (via get_portal_access_key_info) in
            # its response, so an appropriate error can be shown in the UI. We do NOT want to ALWAYS get this
            # information (via get_portal_access_key_info) in the /header endpoint (like in the normal case
            # where there is no error) because it would impact performance (the /header endpoint should be as
            # fast as possible), so we only do it if looks like there is a problem, i.e. as there is here.
            #
            # There is also a separate /portal_access_key endpoint to be called asynchronously by the UI to
            # display any error (e.g. acesss key expired) or warning (e.g. access key expiring soon), but
            # the UI does not use that to redirect to an error page (on error) because, being asynchronous,
            # it could be a jarring UX (i.e. to all of a sudden be redirected to an error page after the
            # current page looks like it is stable), so the UI uses this /portal_access_key endpoint to
            # just display a (red) bar across the top of the page indicating that the accesss key is
            # invalid (e.g. expired) or will expire soon.
            #
            # This is in contrast the the analogous behavior of the Portal SSL certicifcate checking.
            # In that case, since the /header endpoint indirectly and synchronously calls the Portal
            # health endpoint ANYWAYS, we will know at that point if the certificate is problematic,
            # and in that case we can return from the /header endpoint information about the certificate,
            # which the UI can use to redirect to an error, without a jarring UX, i.e. since this the
            # /header endpoint is the first and primary API called by the UI before it can do anything.
            #
            allowed_envs = []
            first_name = None
            last_name = None
            user_exception = True

        authtoken_decoded = {
            "authentication": "auth0",
            "authenticator": authenticator,
            "authenticated_at": jwt_decoded.get("iat"),
            "authenticated_until": jwt_expires_at,
            "authenticated": True,
            "user": email,
            "user_verified": jwt_decoded.get("email_verified"),
            "first_name": first_name,
            "last_name": last_name,
            "allowed_envs": allowed_envs,
            "known_envs": self._envs.get_known_envs(),
            "default_env": self._envs.get_default_env(),
            "domain": domain,
            "site": app.core.get_site_name()
        }
        if user_exception:
            authtoken_decoded["user_exception"] = True
        # JWT-sign-encode the authtoken using our Auth0 client ID (aka audience aka "aud") and
        # secret. This *required* audience is added to the JWT before encoding (done in the
        # jwt_encode function), set to the value we pass here, namely, self._auth0_client;
        # it was also in the given JWT (i.e. jwt_decoded["aud"]), and these should match (no
        # need to check); it came from the original Auth0 invocation on the client-side (in the
        # Auth0Lock box); we communicate it to the client-side via the non-protected /header endpoint.
        return jwt_encode(authtoken_decoded, audience=self._auth0_client, secret=self._auth0_secret)

    def decode_authtoken(self, authtoken: str) -> dict:
        """
        Fully verifies AND decodes and returns the given JWT-signed-encoded authtoken (cookie).
        If not verified (by the jwt_decode function) then None will be returned.
        See create_authtoken (above) for an enumeration of the contents of the authtoken.
        Returns the verified/decoded JWT as a dictionary.
        """
        return jwt_decode(authtoken, self._auth0_client, self._auth0_secret)

    def _create_unauthorized_response(self, request: dict, status: str,
                                      authtoken_decoded: dict, is_authenticated: bool) -> dict:
        """
        Creates a response suitable for a request which is NOT authorized, or NOT authenticated,
        depending on the is_authenticated argument. Note that we still want to return some
        basic info, i.e. the default environment, domain, and aud (Auth0 client ID) is required
        for the Auth0 login box (Auth0Lock) on the client-side (i.e. React UI). This info is
        gotten from the given decoded authtoken or if not set then sets this info explicitly.
        """
        if authtoken_decoded:
            response = authtoken_decoded
        else:
            response = {
                # 2022-10-18
                # No longer including known-envs in unauthorized responses.
                # "known_envs": self._envs.get_known_envs(),
                "default_env": self._envs.get_default_env(),
                "domain": get_request_domain(request),
                "site": "foursight-cgap" if app.core.APP_PACKAGE_NAME == "foursight-cgap" else "foursight-fourfront",
                JWT_AUDIENCE_PROPERTY_NAME: self._auth0_client  # Needed for Auth0Lock login box on client-side.
            }
        response["authenticated"] = is_authenticated
        response["authorized"] = False
        response["status"] = status
        return response

    def _create_unauthenticated_response(self, request: dict, status: str, authtoken_decoded: dict = None) -> dict:
        """
        Creates a response suitable for a request which is NOT authenticated.
        """
        return self._create_unauthorized_response(request, status, authtoken_decoded, is_authenticated=False)

    @function_cache(nocache=None)
    def get_aws_credentials(self, env: str) -> dict:
        """
        Returns basic AWS credentials info (NOT the secret).
        This is just for informational/display purposes in the UI.
        This has nothing to do with the rest of the authentication
        and authorization stuff here but vaguely related so here seems fine.
        """
        aws_credentials = None
        try:
            session = boto3.session.Session()
            credentials = session.get_credentials()
            access_key_id = credentials.access_key
            region_name = session.region_name
            caller_identity = boto3.client("sts").get_caller_identity()
            user_arn = caller_identity["Arn"]
            account_number = caller_identity["Account"]
            aws_credentials = {
                "aws_account_number": account_number,
                "aws_user_arn": user_arn,
                "aws_access_key_id": access_key_id,
                "aws_region": region_name,
                "auth0_client_id": self._auth0_client
            }
            # Try getting the account name though probably no permission at the moment.
            aws_account_name = None
            try:
                aws_credentials["aws_account_name"] = (
                    boto3.client('iam').list_account_aliases()['AccountAliases'][0])
            except Exception as e:
                logger.warning(f"Exception (not fatal) getting AWS account alias: {e}")
            if not aws_account_name:
                try:
                    aws_credentials["aws_account_name"] = (
                        boto3.client('organizations').
                        describe_account(AccountId=account_number).get('Account').get('Name')
                    )
                except Exception as e:
                    logger.warning(f"Exception (not fatal) getting AWS account name: {e}")
        except Exception as e:
            logger.warning(f"Exception (not fatal) getting AWS account info: {e}")
            return None
        return aws_credentials



import re
from typing import Optional


KNOWN_TASK_DEFINITION_OR_SERVICE_MAP = {re.compile(key, re.IGNORECASE): value for key, value in {
    r".*deploy.*initial.*": "deploy_initial",
    r".*deploy.*": "deploy",
    r".*index.*": "indexer",
    r".*ingest.*": "ingester",
    r".*portal.*": "portal",
}.items()}.items()


def get_task_definition_type(task_definition_arn: str) -> Optional[str]:
    return _get_task_definition_or_service_type(task_definition_arn)


def get_service_type(service_arn: str) -> Optional[str]:
    return _get_task_definition_or_service_type(service_arn)


def _get_task_definition_or_service_type(task_definition_or_service_arn: str) -> Optional[str]:
    for pattern, task_definition_or_service_type in KNOWN_TASK_DEFINITION_OR_SERVICE_MAP:
        if pattern.match(task_definition_or_service_arn):
            return task_definition_or_service_type
    return None



# This module defines a "route" decorator that wraps the Chalice route decorator,
# and does authorization (and authentication) checking, tweaks the path appropriately,
# sets up CORS if necessary (only for local development), and has common exception handling.
# We DEFAULT to AUTHORIZATION checking; if not wanted use authorize=False in route decorator.

from chalice import CORSConfig, Response
import logging
from typing import Optional, Tuple
from dcicutils.misc_utils import get_error_message, PRINT
from ...app import app
from ...route_prefixes import CHALICE_LOCAL, ROUTE_PREFIX, ROUTE_EMPTY_PREFIX, ROUTE_PREFIX_EXPLICIT

REACT_API_PATH_COMPONENT = "reactapi"
REACT_UI_PATH_COMPONENT = "react"
ROUTE_PREFIX = ROUTE_PREFIX + ("/" if not ROUTE_PREFIX.endswith("/") else "")

# CORS is need ONLY for local DEVELOPMENT, i.e. when using chalice local!
# Set CORS to True if CHALICE_LOCAL; not needed if running React from Foursight
# directly, on the same port (e.g. 8000), but useful if/when running React on a
# separate port (e.g. 3000) via npm start in foursight-core/react to facilitate
# easy/quick development/changes directly to React UI code.
if CHALICE_LOCAL:
    # Very specific/tricky requirements for running Foursight React UI/API in CORS
    # mode (i.e. UI on localhost:3000 and API on localhost:8000). The allow_origin
    # must be exact (i.e. no "*" allowed), and the allow_credentials must be True.
    # On the client-side (React UI) we must include 'credentials: "include"' in the
    # fetch (if using React fetch, though now we use axios, where we require, in
    # any case, 'withCredentials: "include"').
    _CORS = CORSConfig(
        allow_origin="http://localhost:3000",  # Need this to be explicit not "*"
        allow_credentials=True  # Need this
    )
else:
    _CORS = None

_HTTP_UNAUTHENTICATED = 401
_HTTP_UNAUTHORIZED = 403

logging.basicConfig()
logger = logging.getLogger(__name__)


def route_root() -> Response:
    return app.core.create_redirect_response(f"{ROUTE_PREFIX_EXPLICIT}{REACT_UI_PATH_COMPONENT}")


def route(*args, **kwargs):
    """
    Decorator to wrap the Chalice route decorator to do authentication and authorization
    checking; tweaks the endpoint path appropriately (for API and static files); sets up
    CORS if necessary (for local development); and handles exceptions. Usage looks like:

      @route("/{env}/info", authorize=True)
      def reactapi_route_info(env: str) -> Response:
          return do_route_processing_and_return_response(env)

    Note that functions decorated with this are (if class members) implicitly STATIC methods.
    """
    # Special handling for "root" route, i.e. / (just slash). If NO function specified for the
    # @route (which actually IS allowed) redirect to the main UI route via route_root function
    # above; if a function IS specified for the @route then use that. Complication is for the
    # chalice local case where /api is NOT the AWS lambda enforced base route in which case
    # we support it explicitly for closer compatibility with the normally deployed case.
    if "root" in kwargs:
        if kwargs["root"] is True:
            # Registration for root routes, i.e. / (and /api for chalice local).
            def root_route_registration(wrapped_route_function):
                if not callable(wrapped_route_function):
                    wrapped_route_function = route_root
                PRINT(f"Registering Chalice endpoint: GET {ROUTE_EMPTY_PREFIX} -> {wrapped_route_function.__name__}")
                if ROUTE_EMPTY_PREFIX != "/":
                    # This is true only for the chalice local case.
                    PRINT(f"Registering Chalice endpoint: GET / -> {wrapped_route_function.__name__}")
                    app.route("/", methods=["GET"])(wrapped_route_function)
                return app.route(ROUTE_EMPTY_PREFIX, methods=["GET"])(wrapped_route_function)
            return root_route_registration
        del kwargs["root"]

    if not isinstance(args, Tuple) or len(args) == 0:
        raise Exception("No arguments found for route configuration!")

    path = args[0]

    # This "static" is for serving static files which live in their own specific directory.
    if "static" in kwargs:
        if kwargs["static"] is True:
            path = ROUTE_PREFIX + REACT_UI_PATH_COMPONENT + path
        del kwargs["static"]
    else:
        path = ROUTE_PREFIX + REACT_API_PATH_COMPONENT + path

    # This "define_noenv_route" is for defining a version of the route which does not contain /{env}.
    if "define_noenv_route" in kwargs:
        if kwargs["define_noenv_route"] is True:
            define_noenv_route = True
        del kwargs["define_noenv_route"]
    else:
        define_noenv_route = False

    if path.endswith("/"):
        path = path[:-1]
    if not path.startswith("/"):
        path = "/" + path

    # This "authorize" is to specify whether or not the route required authorization.
    # Note we DEFAULT to AUTHORIZE for the route! The only way to turn it off is to
    # explicitly pass authorize=False to the route decorator.
    if "authorize" in kwargs:
        authorize = kwargs["authorize"] is True
        del kwargs["authorize"]
    else:
        authorize = True

    # As a convenience we allow "method" or "methods" to specify the HTTP verb(s).
    # If no method(s) given then default to GET.
    if "methods" not in kwargs:
        if "method" in kwargs:
            kwargs["methods"] = [kwargs["method"]]
            del kwargs["method"]
        else:
            kwargs["methods"] = ["GET"]
    elif "method" in kwargs:
        del kwargs["method"]

    def route_registration(wrapped_route_function):
        """
        This function is called once for each defined route/endpoint (at app startup).
        """
        def route_function(*args, **kwargs):
            """
            This is the function called on each actual route/endpoint (API) call.
            """
            try:
                env = kwargs.get("env")
                if env and env != "static" and app.core._envs and not app.core._envs.is_known_env(env):
                    env = app.core.get_default_env()
                if authorize:
                    # Note that the "env" argument in the kwargs is the environment name from the endpoint
                    # path; this does NOT have to be present in the endpoint path, BUT if it IS then it
                    # MUST be exactly named "env", otherwise we won't properly do per-env authorization.
                    unauthorized_response = _authorize(app.current_request.to_dict(), env)
                    if unauthorized_response:
                        return unauthorized_response
                # Here we are authenticated and authorized and so we call the actual route function.
                if define_noenv_route and not env:
                    kwargs["env"] = app.core.get_default_env()
                return wrapped_route_function(*args, **kwargs)
            except Exception as e:
                # Common endpoint exception handling here.
                logger.error(f"Exception in route ({wrapped_route_function.__name__}): {get_error_message(e)}")
                return app.core.create_error_response(e)
        if _CORS:
            # Only used for (cross-origin) localhost development (e.g. UI on 3000 and API on 8000).
            kwargs["cors"] = _CORS
        PRINT(f"Registering Chalice endpoint: {' '.join(kwargs['methods'])} {path} -> {wrapped_route_function.__name__}")
        # This is the call that actually registers the Chalice route/endpoint.
        if define_noenv_route:
            noenv_path = path.replace("/{env}", "")
            if noenv_path != path:
                app.route(noenv_path, **kwargs)(route_function)
        return app.route(path, **kwargs)(route_function)
    return route_registration


def _authorize(request: dict, env: Optional[str]) -> Optional[Response]:
    """
    If the given request is UNAUTHORIZED (or UNAUTHENTICATED) then returns
    an appropriate unauthorized (or unauthenticated) response, otherwise None;
    i.e. a return of None means all is well, fully authenticated and authorized.
    """
    authorize_response = app.core.react_authorize(request, env)
    if not authorize_response or not authorize_response["authorized"]:
        # HTTP 401 - Unauthorized (more precisely: Unauthenticated):
        # Request has no or invalid credentials.
        # HTTP 403 - Forbidden (more precisely: Unauthorized):
        # Request has valid credentials but no privileges for resource.
        http_status = _HTTP_UNAUTHENTICATED if not authorize_response["authenticated"] else _HTTP_UNAUTHORIZED
        return app.core.create_response(http_status=http_status, body=authorize_response)
    return None



from .helpers.confchecks import (
    check_function, CheckResult,
)
from dcicutils.ecs_utils import ECSUtils
from dcicutils.ecr_scripts import DEFAULT_ACCOUNT_NUMBER, ecr_command_context, ECRCommandContext


@check_function(env_name=None, account_number=None, ecr_repo=None, image_tag=None)
def rollback_application_version(connection, **kwargs):
    """ Checks for an active deployment that has been running for > 30 minutes,
        after which we can assume there is an issue with it and it should be
        rolled back. In our setup, rolling back means moving the image tag back
        to the previous version. If that version fails to deploy, this check will
        run again, continuing the rollback mechanism until a stable version is found.

        Note that this check assumes it is run on a 15 minute schedule - it will not function
        correctly otherwise! It also will ignore deployments that occur while counts are uneven.
    """
    check = CheckResult(connection, 'rollback_application_version')
    env = kwargs.get('env_name') or connection.ff_env
    ecr_repo = kwargs.get('ecr_repo', 'main')  # typically main
    image_tag = kwargs.get('image_tag', 'latest')  # typically 'latest' but could be blue/green
    account_number = kwargs.get('account_number', DEFAULT_ACCOUNT_NUMBER)  # os.environ.get('ACCOUNT_NUMBER')
    ecs_utils = ECSUtils()
    clusters = ecs_utils.list_ecs_clusters()
    matching_cluster = list(filter(lambda c: env in c, clusters))
    if len(matching_cluster) == 0:
        check.status = 'FAIL'
        check.summary = f'Could not resolve a cluster given env name {env}'
        check.brief_output = check.full_output = check.summary
        return check
    elif len(matching_cluster) > 1:
        check.status = 'FAIL'
        check.summary = f'Could not resolve a single cluster given env name {env}, resolved:' \
                        f' {matching_cluster}'
        check.brief_output = check.full_output = check.summary
        return check
    else:  # we found a single cluster to check
        matching_cluster = matching_cluster[0]
        services = ecs_utils.list_ecs_services(cluster_name=matching_cluster)
        deploy_is_active = ecs_utils.service_has_active_deployment(cluster_name=matching_cluster, services=services)
        # check previous run, checking that ES is reachable
        if deploy_is_active and (connection.test_es_connection()):
            last_result = check.get_primary_result()
            prior_result = check.get_closest_result(diff_mins=30)
            # set to warn if either of previous results don't exist
            if not last_result or not prior_result:
                check.status = 'WARN'
                check.summary = f'Detected that deployment is still running - after 3 consecutive checks, image will' \
                                f' be rolled back on cluster {matching_cluster}'
                check.brief_output = check.full_output = check.summary
                return check
            # 2 consecutive prior warns indicate deployment failure
            elif last_result.status == 'WARN' and prior_result.status == 'WARN':
                with ecr_command_context(account_number=account_number, ecs_repository=ecr_repo):
                    ecr = ECRCommandContext(account_number, ecs_repository=ecr_repo)
                    images = ecr.get_images_descriptions()
                    for i in range(len(images) - 1):
                        current_image = images[i]
                        if image_tag in current_image.get('imageTags', []):
                            image_prior_to_latest = images[i+1]
                            ecr.add_image_tag(digest=image_prior_to_latest['imageDigest'], tag=image_tag)
                check.status = 'PASS'
                check.summary = f'Triggered image rollback on tag {image_tag} on repo {ecr_repo} associated with' \
                                f' cluster {matching_cluster}'
                check.brief_output = check.full_output = check.summary
                return check
            else:  # we are at the first or second warn
                check.status = 'WARN'
                check.summary = f'Detected that deployment is still running - after 3 consecutive checks, image will' \
                                f' be rolled back on cluster {matching_cluster}'
                check.brief_output = check.full_output = check.summary
                return check
        else:
            check.status = 'PASS'
            check.summary = 'No active deployment detected (or counts are uneven) - all good!'
            check.brief_output = check.full_output = check.summary
            return check



import base64
import json
from typing import List, Optional, Tuple, Union
import uuid
from ...boto_s3 import boto_s3_client
from .datetime_utils import convert_datetime_to_utc_datetime_string


def read_ingestion_submissions(bucket: str,
                               offset: int = 0, limit: int = -1,
                               sort: Optional[str] = None) -> dict:
    """
    Returns information about the ingestion submission keys within the metadata-bundles-bucket,
    This keys in this bucket are organized by ingestion submission uuid like this:

    -> 9f9e5033-a1bc-4e72-ad0d-1873d4253f25/started.txt
    -> 9f9e5033-a1bc-4e72-ad0d-1873d4253f25/submission.json

    So we return a list (actually a dictionary with the list together with paging info)
    of objects by uuid like this:

       {
           "uuid": 9f9e5033-a1bc-4e72-ad0d-1873d4253f25/started.txt
           "started": true
           "done": true
           "modified": "2023-09-15 13:45:20 UTC"
       }

    These are ordered, by default, by reverse modified datetime (most recent first).

    Some ingestion info can be gotten via Portal but only that which is written to the database,
    which does not include the details of the ingested submission, i.e. the actual list of uuids
    for objects which were created, update, or validated; nor does it include any details of any
    validation errors (or other errors or exceptions) which may have occurred.
    """
    s3 = boto_s3_client()

    def is_uuid(value: str) -> bool:
        try:
            return str(uuid.UUID(value)).lower() == value.lower()
        except Exception as e:
            return False

    def get_uuid_and_file_from_key(key: str) -> Tuple[Optional[str], Optional[str]]:
        parts = key.split("/")
        if len(parts) != 2 or not is_uuid(parts[0]) or not parts[1]:
            return (None, None)
        return (parts[0], parts[1])

    def add_key_to_keys(key: dict, key_size: int, keys: dict) -> None:
        key_uuid = key["uuid"]
        key_file = key["file"]
        existing_key = keys.get(key_uuid)
        if existing_key:
            # We take the most recently modified date within the
            # bucket amont the keys with the same (prefix) uuid.
            if key["modified"] > existing_key["modified"]:
                existing_key["modified"] = key["modified"]
            if key_file.startswith("datafile"):
                existing_key["file"] = key_file
            key = existing_key
        else:
            keys[key_uuid] = key
        if _is_started_file(key_file):
            key["started"] = True
        if _is_done_file(key_file):
            key["done"] = True
        if _is_error_file(key_file):
            key["error"] = True
        if not key.get("files"):
            key["files"] = []
        if not [file for file in key["files"] if file["file"] == key_file]:
            file = {"file": key_file, "modified": key["modified"], "size": key_size}
            if _is_detail_file(key_file):
                file["detail"] = True
            key["files"].append(file)
        if key["file"] and not key["file"].startswith("datafile"):
            key["file"] = None

    # Unfortunately AWS/boto3 does not allow sorting, so we need to read all the keys in the bucket;
    # this may be slow, but it's quicker than manually rummaging around when you need to find something.

    def get_chunk_of_keys(bucket: str, next_token: Optional[str] = None) -> Tuple[Optional[List[dict]], Optional[str]]:
        keys = {}
        MAX_KEYS_PER_READ = 1000
        if next_token:
            response = s3.list_objects_v2(Bucket=bucket, MaxKeys=MAX_KEYS_PER_READ, ContinuationToken=next_token)
        else:
            response = s3.list_objects_v2(Bucket=bucket, MaxKeys=MAX_KEYS_PER_READ)
        for item in response.get("Contents", []):
            key = item["Key"]
            uuid, file = get_uuid_and_file_from_key(key)
            if not uuid or not file:
                continue
            key = {"uuid": uuid, "file": file, "modified": item["LastModified"]}
            add_key_to_keys(key, item["Size"], keys)
        return (keys, response.get("NextContinuationToken"))

    keys = {}
    next_token = None
    while True:
        chunk_of_keys, next_token = get_chunk_of_keys(bucket, next_token)
        if chunk_of_keys:
           for key in chunk_of_keys:
               keys[key] = chunk_of_keys[key]
        if not next_token:
            break
    keys = [keys[key] for key in keys]
    total = len(keys)

    sort_key = None
    if sort:
        if sort.endswith(".desc"):
            sort_key = sort[:-5].lower()
            sort_reverse = True
        elif sort.endswith(".asc"):
            sort_key = sort[:-4].lower()
            sort_reverse = False
    if sort_key in ["uuid", "modified"]:
        keys = sorted(keys, key=lambda key: key[sort_key], reverse=sort_reverse)

    offset = (total if offset > total else offset) if offset >= 0 else 0
    limit = limit if limit >= 0 else total
    end_offset = offset + limit if offset + limit <= total else total
    keys = keys[offset:end_offset]

    # Convert datatime objects to strings, here just for minor performance reason,
    # rather than doing it for all keys, just do the ones we are returning.
    for key in keys:
        key["modified"] = convert_datetime_to_utc_datetime_string(key["modified"])
        for file in key["files"]:
            file["modified"] = convert_datetime_to_utc_datetime_string(file["modified"])

    return {
        "paging": {
            "total": total, "count": end_offset - offset,
            "limit": limit, "offset": offset, "more": max(total - offset - limit, 0)
        },
        "list": keys
    }


def read_ingestion_submission_detail(bucket: str, uuid: str) -> Optional[str]:
    file = f"submission.json"
    contents = _read_s3_key(bucket, f"{uuid}/{file}", is_json=True)
    if not contents:
        alternative_file = f"submission.json.json"
        contents = _read_s3_key(bucket, f"{uuid}/{alternative_file}", is_json=True)
        if contents:
            file = alternative_file
    return {"file": file, "detail": contents}


def read_ingestion_submission_summary(bucket: str, uuid: str) -> Optional[str]:
    file = f"summary.json"
    contents = _read_s3_key(bucket, f"{uuid}/{file}", is_json=True)
    return {"file": file, "summary": contents, "started": _exists_s3_key(bucket, file)}


def read_ingestion_submission_manifest(bucket: str, uuid: str) -> Optional[str]:
    file = f"manifest.json"
    return {"file": file, "manifest": _read_s3_key(bucket, f"{uuid}/{file}", is_json=True)}


def read_ingestion_submission_upload_info(bucket: str, uuid: str) -> Optional[dict]:
    file = f"upload_info.txt"
    return {"file": file, "upload_info": _read_s3_key(bucket, f"{uuid}/{file}", is_json=True)}


def read_ingestion_submission_resolution(bucket: str, uuid: str) -> Optional[str]:
    file = f"resolution.json"
    return {"file": file, "resolution": _read_s3_key(bucket, f"{uuid}/{file}", is_json=True)}


def read_ingestion_submission_submission_response(bucket: str, uuid: str) -> Optional[str]:
    file = f"submission_response.txt"
    return {"file": file, "submission_response": _read_s3_key(bucket, f"{uuid}/{file}", is_json=False)}


def read_ingestion_submission_traceback(bucket: str, uuid: str) -> Optional[str]:
    file = f"traceback.txt"
    return {"file": file, "traceback": _read_s3_key(bucket, f"{uuid}/{file}", is_json=False)}


def read_ingestion_submission_validation_report(bucket: str, uuid: str) -> Optional[str]:
    file = f"validation_report.txt"
    return {"file": file, "validation_report": _read_s3_key(bucket, f"{uuid}/{file}", is_json=False)}


def _read_s3_key(bucket: str, key: str,
                 is_json: bool = False, is_binary: bool = False) -> Optional[Union[dict, str, bytes]]:
    if is_json:
        is_binary = False
    try:
        contents = boto_s3_client().get_object(Bucket=bucket, Key=key)["Body"].read()
        if not is_binary:
            contents = contents.decode("utf-8")
            if is_json:
                contents = json.loads(contents)
        else:
            contents = base64.b64encode(contents).decode("utf-8")
        return contents
    except Exception:
        return None


def _exists_s3_key(bucket: str, key: str) -> Optional[str]:
    try:
        return boto_s3_client().head_object(Bucket=bucket, Key=key) is not None
    except Exception:
        return False


def _sizeof_s3_key(bucket: str, key: str) -> Optional[int]:
    try:
        return boto_s3_client().head_object(Bucket=bucket, Key=key).get("ContentLength")
    except Exception:
        return None


def _is_started_file(file: str) -> bool:
    return file == "started.txt"


def _is_done_file(file: str) -> bool:
    return _is_detail_file(file)


def _is_error_file(file: str) -> bool:
    return file == "traceback.txt"


def _is_detail_file(file: str) -> bool:
    # We note this file in particular because it could be very large;
    # it contains an itemization (by uuid) of every object processed.
    # Files like submission.json.json exist; probably a (hopefully past) mistake; but handle it.
    return file == "submission.json" or file == "submission.json.json"
